@@ -11,38 +11,46 @@ which may be used as a catch-all filesystem exception.
 from __future__ import unicode_literals
 from __future__ import print_function
 
+import functools
+import typing
+
 import six
 from six import text_type
 
+if False:  # typing.TYPE_CHECKING
+    from typing import Optional, Text
+
+
 __all__ = [
-    'CreateFailed',
-    'DestinationExists',
-    'DirectoryExists',
-    'DirectoryExpected',
-    'DirectoryNotEmpty',
-    'FileExists',
-    'FileExpected',
-    'FilesystemClosed',
-    'FSError',
-    'IllegalBackReference',
-    'InsufficientStorage',
-    'InvalidCharsInPath',
-    'InvalidPath',
-    'MissingInfoNamespace',
-    'NoSysPath',
-    'NoURL',
-    'OperationFailed',
-    'OperationTimeout',
-    'PathError',
-    'PermissionDenied',
-    'RemoteConnectionError',
-    'RemoveRootError',
-    'ResourceError',
-    'ResourceInvalid',
-    'ResourceLocked',
-    'ResourceNotFound',
-    'ResourceReadOnly',
-    'Unsupported',
+    "BulkCopyFailed",
+    "CreateFailed",
+    "DestinationExists",
+    "DirectoryExists",
+    "DirectoryExpected",
+    "DirectoryNotEmpty",
+    "FileExists",
+    "FileExpected",
+    "FilesystemClosed",
+    "FSError",
+    "IllegalBackReference",
+    "InsufficientStorage",
+    "InvalidCharsInPath",
+    "InvalidPath",
+    "MissingInfoNamespace",
+    "NoSysPath",
+    "NoURL",
+    "OperationFailed",
+    "OperationTimeout",
+    "PathError",
+    "PermissionDenied",
+    "RemoteConnectionError",
+    "RemoveRootError",
+    "ResourceError",
+    "ResourceInvalid",
+    "ResourceLocked",
+    "ResourceNotFound",
+    "ResourceReadOnly",
+    "Unsupported",
 ]
 
 
@@ -51,10 +59,9 @@ class MissingInfoNamespace(AttributeError):
     """
 
     def __init__(self, namespace):
+        # type: (Text) -> None
         msg = "namespace '{}' is required for this attribute"
-        super(MissingInfoNamespace, self).__init__(
-            msg.format(namespace)
-        )
+        super(MissingInfoNamespace, self).__init__(msg.format(namespace))
 
 
 @six.python_2_unicode_compatible
@@ -65,16 +72,19 @@ class FSError(Exception):
     default_message = "Unspecified error"
 
     def __init__(self, msg=None):
+        # type: (Optional[Text]) -> None
         self._msg = msg or self.default_message
         super(FSError, self).__init__()
 
     def __str__(self):
+        # type: () -> Text
         """Return the error message.
         """
         msg = self._msg.format(**self.__dict__)
         return msg
 
     def __repr__(self):
+        # type: () -> Text
         msg = self._msg.format(**self.__dict__)
         return "{}({!r})".format(self.__class__.__name__, msg)
 
@@ -86,11 +96,42 @@ class FilesystemClosed(FSError):
     default_message = "attempt to use closed filesystem"
 
 
+class BulkCopyFailed(FSError):
+    """A copy operation failed in worker threads."""
+
+    default_message = (
+        "One or more copy operations failed (see errors attribute)"
+    )
+
+    def __init__(self, errors):
+        self.errors = errors
+        super(BulkCopyFailed, self).__init__()
+
+
 class CreateFailed(FSError):
     """Filesystem could not be created.
     """
 
-    default_message = "unable to create filesystem"
+    default_message = "unable to create filesystem, {details}"
+
+    def __init__(self, msg=None, exc=None):
+        # type: (Optional[Text], Optional[Exception]) -> None
+        self._msg = msg or self.default_message
+        self.details = "" if exc is None else text_type(exc)
+        self.exc = exc
+
+    @classmethod
+    def catch_all(cls, func):
+        @functools.wraps(func)
+        def new_func(*args, **kwargs):
+            try:
+                return func(*args, **kwargs)
+            except cls:
+                raise
+            except Exception as e:
+                raise cls(exc=e)
+
+        return new_func  # type: ignore
 
 
 class PathError(FSError):
@@ -100,6 +141,7 @@ class PathError(FSError):
     default_message = "path '{path}' is invalid"
 
     def __init__(self, path, msg=None):
+        # type: (Text, Optional[Text]) -> None
         self.path = path
         super(PathError, self).__init__(msg=msg)
 
@@ -118,6 +160,7 @@ class NoURL(PathError):
     default_message = "path '{path}' has no '{purpose}' URL"
 
     def __init__(self, path, purpose, msg=None):
+        # type: (Text, Text, Optional[Text]) -> None
         self.purpose = purpose
         super(NoURL, self).__init__(path, msg=msg)
 
@@ -142,10 +185,16 @@ class OperationFailed(FSError):
 
     default_message = "operation failed, {details}"
 
-    def __init__(self, path=None, exc=None, msg=None):
+    def __init__(
+        self,
+        path=None,  # type: Optional[Text]
+        exc=None,  # type: Optional[Exception]
+        msg=None,  # type: Optional[Text]
+    ):
+        # type: (...) -> None
         self.path = path
         self.exc = exc
-        self.details = '' if exc is None else text_type(exc)
+        self.details = "" if exc is None else text_type(exc)
         self.errno = getattr(exc, "errno", None)
         super(OperationFailed, self).__init__(msg=msg)
 
@@ -199,6 +248,7 @@ class ResourceError(FSError):
     default_message = "failed on path {path}"
 
     def __init__(self, path, exc=None, msg=None):
+        # type: (Text, Optional[Exception], Optional[Text]) -> None
         self.path = path
         self.exc = exc
         super(ResourceError, self).__init__(msg=msg)
@@ -288,7 +338,7 @@ class IllegalBackReference(ValueError):
     """
 
     def __init__(self, path):
-        _msg = \
-            "path '{path}' contains back-references outside of filesystem"
+        # type: (Text) -> None
+        _msg = "path '{path}' contains back-references outside of filesystem"
         _msg = _msg.format(path=path)
         super(IllegalBackReference, self).__init__(_msg)
